# function defined in this s-file:
#    quad2  : integrator of f(x) over (a,b).
#    (C) Shin'ichi OISHI   ---2001/8/25---
#
#    This is originally made by Ooura as a C program.
#    Oishi has translated it into a s-file.
#
#    Numerical Automatic Integrator for Improper Integral
#    method    : Double Exponential (DE) Transformation
#    dimension : one
#    [description]
#        I = integral of f(x) over (a,b)
#    [usage]
#        quad2(a, b, eps)
#    [parameters]
#        f         : integrand f(x) (double (*f)(double))
#        a         : lower limit of integration (double)
#        b         : upper limit of integration (double)
#        eps       : relative error requested (double)
#    [answer]
#        i         : approximation to the integral
#        err       : estimate of the absolute error
#    [remarks]
#        function
#            f(x) needs to be analytic over (a,b).
#        relative error
#            eps is relative error requested excluding 
#            cancellation of significant digits.
#            i.e. eps means : (absolute error) / 
#                             (integral_a^b |f(x)| dx).
#            eps does not mean : (absolute error) / I.
#        error message
#            err >= 0 : normal termination.
#            err < 0  : abnormal termination (m >= mmax).
#                       i.e. convergent error is detected :
#                           1. f(x) or (d/dx)^n f(x) has 
#                              discontinuous points or sharp 
#                              peaks over (a,b).
#                              you must divide the interval 
#                              (a,b) at this points.
#                           2. relative error of f(x) is 
#                              greater than eps.
#                           3. f(x) has oscillatory factor 
#                              and frequency of the oscillation 
#                              is very high.

function quad2(f, a, b, eps) {
    mmax = 256;
    efs = 0.1;
    hoff = 8.5;
    pi2 = 2 * atan(1.0);
    epsln = 1 - log(efs * eps);
    epsh = sqrt(efs * eps);
    h0 = hoff / epsln;
    ehp = exp(h0);
    ehm = 1 / ehp;
    epst = exp(-ehm * epsln);
    ba = b - a;
    ir = eval(f|(a + b) * 0.5) * (ba * 0.25);
    i = ir * (2 * pi2);
    err = fabs(i) * epst;
    h = 2 * h0;
    m = 1;
    do {
        iback = i;
        irback = ir;
        t = h * 0.5;
        do {
            em = exp(t);
            ep = pi2 * em;
            em = pi2 / em;
            do {
                xw = 1 / (1 + exp(ep - em));
                xa = ba * xw;
                wg = xa * (1 - xw);
                fa = eval(f|a + xa) * wg;
                fb = eval(f|b - xa) * wg;
                ir = ir + fa + fb;
                i = i + (fa + fb) * (ep + em);
                errt = (fabs(fa) + fabs(fb)) * (ep + em);
                if (m == 1) {err = err + errt * epst};
                ep = ep * ehp;
                em = em * ehm
            } while ((errt > err) || (xw > epsh));
            t = t + h
        } while (t < h0);
        if (m == 1) {
            errh = (err / epst) * epsh * h0;
            errd = 1 + 2 * errh
        } else {
            errd = h * (fabs(i - 2 * iback) + 4 * fabs(ir - 2 * irback))
        };
        h = h * 0.5;
        m = m * 2
    } while ((errd > errh) && (m < mmax));
    i = i * h;
    if (errd > errh) {
        err = -errd * m
    } else {
        err = errh * epsh * m / (2 * efs)
    };
    <i, err>
}
